import { Action } from './action.js';
import { Actor } from './actor.js';
import { CryptoUtil } from './crypto.js';
import { PostContent } from './post-content.js';

class Request {
  constructor(fee, content) {
    this.id = CryptoUtil.randomUUID();
    this.fee = fee;
    this.content = content;
    this.worker = null;
  }
}

/**
 * Purpose: Enable fee-driven work requests, to be completed by workers from the availability pool
 */
export class Business extends Actor {
  constructor(bench, forum, availability, name, scene) {
    super(name, scene);
    this.bench = bench;
    this.forum = forum;
    this.availability = availability;

    this.actions = {
      assignWork: new Action('assign work', scene),
      submitPost: new Action('submit post', scene),
      initiateValidationPool: new Action('initiate validation pool', scene),
    };

    this.requests = new Map();
  }

  async submitRequest(fee, content) {
    const request = new Request(fee, content);
    this.requests.set(request.id, request);
    await this.actions.assignWork.log(this, this.availability);
    const worker = await this.availability.assignWork(request.id);
    request.worker = worker;
    return request.id;
  }

  async getRequest(requestId) {
    const request = this.requests.get(requestId);
    return request;
  }

  async submitWork(reputationPublicKey, requestId, workEvidence, { tokenLossRatio, duration }) {
    const request = this.requests.get(requestId);
    if (!request) {
      throw new Error(`Request not found! id: ${requestId}`);
    }

    if (reputationPublicKey !== request.worker.reputationPublicKey) {
      throw new Error('Work evidence must be submitted by the assigned worker!');
    }

    // Create a post representing this submission.
    const post = new PostContent({
      requestId,
      workEvidence,
    });
    const requestIndex = Array.from(this.requests.values())
      .findIndex(({ id }) => id === request.id);
    post.setTitle(`Work Evidence ${requestIndex + 1}`);

    await this.actions.submitPost.log(this, this.forum);
    const postId = await this.forum.addPost(reputationPublicKey, post);

    // Initiate a validation pool for this work evidence.
    await this.actions.initiateValidationPool.log(this, this.bench);
    const pool = await this.bench.initiateValidationPool({
      postId,
      fee: request.fee,
      duration,
      tokenLossRatio,
    }, {
      reputationPublicKey,
      authorStakeAmount: request.worker.stakeAmount,
      tokenId: request.worker.tokenId,
    });

    // When the validation pool concludes,
    // reputation should be awarded and fees should be distributed.
    return pool;
  }
}
