import { Actor } from './actor.js';
import { WDAG } from './wdag.js';
import { Action } from './action.js';
import { CryptoUtil } from './crypto.js';
import params from '../params.js';
import { ReputationHolder } from './reputation-holder.js';
import { displayNumber } from '../util.js';

const CITATION = 'citation';
const BALANCE = 'balance';

class Post extends Actor {
  constructor(forum, authorPublicKey, postContent) {
    const index = forum.posts.countVertices();
    const name = `Post${index + 1}`;
    super(name, forum.scene);
    this.id = postContent.id ?? `post_${CryptoUtil.randomUUID().slice(0, 4)}`;
    this.authorPublicKey = authorPublicKey;
    this.value = 0;
    this.initialValue = 0;
    this.citations = postContent.citations;
    this.title = postContent.title;
    const revaluationTotal = this.citations.reduce((total, { weight }) => total += Math.abs(weight), 0);
    if (revaluationTotal > params.revaluationLimit) {
      throw new Error('Post revaluation total exceeds revaluation limit '
      + `(${revaluationTotal} > ${params.revaluationLimit})`);
    }
    if (this.citations.some(({ weight }) => Math.abs(weight) > 1)) {
      throw new Error('Each citation weight must be in the range [-1, 1]');
    }
  }

  getLabel() {
    return `${this.name}
    <table><tr>
      <td>initial</td>
      <td>${displayNumber(this.initialValue)}</td>
    </tr><tr>
      <td>value</td>
      <td>${displayNumber(this.value)}</td>
    </tr></table>`
      .replaceAll(/\n\s*/g, '');
  }
}

/**
 * Purpose: Maintain a directed, acyclic, weighted graph of posts referencing other posts
 */
export class Forum extends ReputationHolder {
  constructor(name, scene) {
    super(`forum_${CryptoUtil.randomUUID()}`, name, scene);
    this.id = this.reputationPublicKey;
    this.posts = new WDAG(scene);
    this.actions = {
      addPost: new Action('add post', scene),
      propagateValue: new Action('propagate value', this.scene),
      transfer: new Action('transfer', this.scene),
    };
  }

  async addPost(authorId, postContent) {
    const post = new Post(this, authorId, postContent);
    await this.actions.addPost.log(this, post);
    this.posts.addVertex(post.id, post, post.getLabel());
    for (const { postId: citedPostId, weight } of post.citations) {
      this.posts.addEdge(CITATION, post.id, citedPostId, weight);
    }
    return post.id;
  }

  getPost(postId) {
    return this.posts.getVertexData(postId);
  }

  getPosts() {
    return this.posts.getVerticesData();
  }

  async setPostValue(post, value) {
    post.value = value;
    await post.setValue('value', value);
    this.posts.setVertexLabel(post.id, post.getLabel());
  }

  getTotalValue() {
    return this.getPosts().reduce((total, { value }) => total += value, 0);
  }

  async onValidate({
    bench, postId, tokenId,
  }) {
    this.activate();
    const initialValue = bench.reputation.valueOf(tokenId);
    const postVertex = this.posts.getVertex(postId);
    const post = postVertex.data;
    post.setStatus('Validated');
    post.initialValue = initialValue;
    this.posts.setVertexLabel(post.id, post.getLabel());

    // Store a reference to the reputation token associated with this post,
    // so that its value can be updated by future validated posts.
    post.tokenId = tokenId;

    const rewardsAccumulator = new Map();

    // Compute rewards
    await this.propagateValue(
      { to: postVertex, from: { data: this } },
      { rewardsAccumulator, increment: initialValue },
    );

    // Apply computed rewards to update values of tokens
    for (const [id, value] of rewardsAccumulator) {
      if (value < 0) {
        bench.reputation.transferValueFrom(id, post.tokenId, -value);
      } else {
        bench.reputation.transferValueFrom(post.tokenId, id, value);
      }
    }

    // Transfer ownership of the minted/staked token, from the forum to the post author
    bench.reputation.transferFrom(this.id, post.authorPublicKey, post.tokenId);
    const toActor = this.scene.findActor((actor) => actor.reputationPublicKey === post.authorPublicKey);
    const value = bench.reputation.valueOf(post.tokenId);
    this.actions.transfer.log(this, toActor, `(value: ${value})`);
    this.deactivate();
  }

  /**
   * @param {Edge} edge
   * @param {Object} opaqueData
   */
  async propagateValue(edge, { rewardsAccumulator, increment, depth = 0 }) {
    const postVertex = edge.to;
    const post = postVertex?.data;

    this.actions.propagateValue.log(edge.from.data, post, `(${increment})`);

    console.log('propagateValue start', {
      from: edge.from.id,
      to: edge.to.id,
      depth,
      value: post.value,
      increment,
    });

    if (params.referenceChainLimit === null || depth <= params.referenceChainLimit) {
      for (const citationEdge of this.posts.getEdges(CITATION, post)) {
        const { to: citedPostVertex, data: { weight } } = citationEdge;
        const citedPost = citedPostVertex.data;
        let outboundAmount = weight * increment;
        const balance = this.posts.getEdge(BALANCE, postVertex, citedPostVertex)?.data || 0;
        console.log('Citation', {
          citationEdge, outboundAmount, balance, citedPostValue: citedPost.value,
        });
        // We need to ensure that we propagate no more reputation than we leached
        if (outboundAmount < 0) {
          outboundAmount = Math.max(outboundAmount, -citedPost.value);
          if (depth > 0) {
            outboundAmount = Math.max(outboundAmount, -balance);
          }
        }
        increment -= outboundAmount * params.leachingValue;
        this.posts.setEdge(BALANCE, postVertex, citedPostVertex, balance + outboundAmount);
        await this.propagateValue(citationEdge, {
          rewardsAccumulator,
          increment: outboundAmount,
          depth: depth + 1,
        });
      }
    }

    const newValue = post.value + increment;

    console.log('propagateValue end', {
      depth,
      increment,
      newValue,
    });

    // Award reputation to post author
    rewardsAccumulator.set(post.tokenId, increment);

    // Increment the value of the post
    await this.setPostValue(post, newValue);
  }
}
